(* Time-stamp: <modified the 01/07/2014 (at 09:43) by Erwan Jahier> *)

open Soc2cIdent

let (mem_interface : Soc.t -> string -> bool) =
  fun soc id -> 
    let ins,outs = soc.profile in
    List.mem_assoc id ins || List.mem_assoc id outs

let rec (string_of_var_expr: Soc.t -> Soc.var_expr -> string) = 
  fun soc -> function
    | Const("true", _) -> "_true"
    | Const("false", _) -> "_false"
    | Const(id, _) -> id2s id
    | Var ("_memory",_)   -> (* Clutch! it's not an interface var... *) "ctx->_memory" 
    | Var (id,_)   -> 
      if not (mem_interface soc id) then id2s id 
      else if SocUtils.is_memory_less soc then
          Printf.sprintf "%s.%s" (get_ctx_name soc.key) (id2s id)
        else 
          Printf.sprintf "ctx->%s" (id2s id)
    | Field(f, id,_) -> Printf.sprintf "%s.%s" (string_of_var_expr soc f) (id2s id) 
    | Index(f, index,_) -> Printf.sprintf "%s[%i]" (string_of_var_expr soc f) index
    | Slice(f,fi,la,st,wi,vt) -> assert false (* should not occur *)


(* exported *) 
let rec (gen_assign : Data.t  -> string -> string -> string -> string) =
  fun t vi vo size -> 
    match t with
      | Alias(_,t) -> gen_assign t vi vo size
      | Enum _  
      | Struct(_) (* should I rather use memcpy for struct? *)
      | Bool | Int | Real -> 
        Printf.sprintf "  %s = %s;\n" vi vo
      | Alpha(_) (* dead code ? *) 
      | Array(_) -> 
        Printf.sprintf "  memcpy(%s, %s, %s);\n" vi vo size

      | Extern (id) -> 
        (* what should i do for extern types? Ask the user to provide the
           copy function I guess *)
        Printf.sprintf "  cpy_%s(%s, %s);\n" id vi vo 

let (gen_assign_var_expr : Soc.t -> Soc.var_expr -> Soc.var_expr -> string) =
fun soc vi vo -> 
  match vi,vo with
    | Slice _, _ -> assert false
    | _, Slice _ ->  assert false
    | _,_ -> 
      gen_assign (Soc.data_type_of_var_expr vi)
        (string_of_var_expr soc vi) (string_of_var_expr soc vo)
        (Printf.sprintf "sizeof(%s)" (string_of_var_expr soc vo))

let id2s = Soc2cIdent.id2s

let (step_name : Soc.key -> string -> string) =
  fun sk sm -> 
    let str = Printf.sprintf "%s_%s" (Soc2cIdent.get_soc_name sk) sm in
    id2s str

type var_kind = (* XXX poor names: fixme! *)
  | ML_IO of Soc.key (* for idents that are part of a MemoryLess soc interface *)
  | M_IO (* for idents that are part of a soc interface with Memories*)
  | Local (* for soc local variables *)

let (ctx_var : var_kind -> Ident.t -> string) =
  fun opt id -> 
    match opt with
      | ML_IO sk -> Printf.sprintf "%s_ctx.%s" (Soc2cIdent.get_soc_name sk) (id2s id)
      | M_IO  ->  Printf.sprintf "ctx->%s" (id2s id)
      | Local -> Printf.sprintf "%s" (id2s id)

let (list_split : 'a list -> int -> 'a list * 'a list) =
  fun l s ->
    let rec aux s l acc =
      match s,l with
        | 0, _ -> List.rev acc,l
        | _, x::l -> aux (s-1) l (x::acc)
        | _, [] -> assert false
    in 
    aux s l []

let _ = assert (list_split [1;2;3;4;5;6] 3 = ([1;2;3],[4;5;6]))


(* exported *) 
let (gen_step_call : Soc.t -> Soc.t -> Soc.var_expr list -> Soc.var_expr list -> 
     string -> string -> string -> string) =
  fun soc called_soc vel_out vel_in ctx sname step_arg -> 
    let called_soc_name,_,_ = called_soc.key in 
    if called_soc_name = "Lustre::if" then
      let c,vel_in= match vel_in with [] -> assert false | c::l -> c,l in 
      let s = (List.length vel_out) in
      let vel_in_t, vel_in_e = list_split vel_in s in
      let lt = List.map2 (gen_assign_var_expr soc) vel_out vel_in_t in
      let le = List.map2 (gen_assign_var_expr soc) vel_out vel_in_e in
      "   if ("^(string_of_var_expr soc c)  ^ " == _true) {\n   "^
        (String.concat "   " lt)^ "   } else {\n   "^
        (String.concat "   " le)^ "   }\n"
    else
      let vel_in = List.map (string_of_var_expr soc) vel_in in
      let vel_out = List.map (string_of_var_expr soc) vel_out in
      let si_str =
        if vel_in = [] then "" (* occurs for pre *) else 
          let inputs = fst called_soc.profile in
          let l = try (
            List.map2 (fun (name, t) ve -> 
              gen_assign t (Printf.sprintf "%s.%s" ctx name) ve (Printf.sprintf "sizeof(%s)" ve)) 
              inputs vel_in
          ) with _ -> assert false (* are all parameters necessaryly used? *)
          in
          (String.concat "" l) 
      in
      let so_str =
        if vel_out = [] then "" (* occurs for pre *) else 
          let outputs = snd called_soc.profile in
          let l = try (
            List.map2
              (fun  (name,t) ve -> 
                let ve2 = Printf.sprintf "%s.%s" ctx name in
                gen_assign t ve ve2 (Printf.sprintf "sizeof(%s)" ve2)) outputs vel_out
          ) with _ -> assert false
          in
          (String.concat "" l) ^"\n"
      in
      let str = Printf.sprintf "  %s(%s); \n" (step_name called_soc.key sname) step_arg in
      (si_str ^ str ^ so_str)



