(* Time-stamp: <modified the 01/07/2014 (at 14:23) by Erwan Jahier> *)

open Data
open Soc
open Soc2cIdent

(* A boring but simple module... *)

(* XXX should i use gen_assign here? for the time being, its useless as
   there is no binop (nor unop) over arrays.
*)
let (lustre_binop : Soc.key -> string -> string) =
  fun sk op  -> 
    let ctx = get_ctx_name sk in
    Printf.sprintf "  %s.z = (%s.x %s %s.y);\n" ctx ctx op ctx

let (lustre_unop : Soc.key -> string -> string) =
  fun sk op  -> 
    let ctx = get_ctx_name sk in
    (* use gen_assign? *)
    Printf.sprintf"  %s.z = %s %s.x;\n" ctx op ctx

let (lustre_ite : Soc.key -> string) =
  fun sk -> 
    let ctx = get_ctx_name sk in
    let (_,_::t::_,_) = sk in
(*     Printf.sprintf"  %s.z = (%s.c)? %s.xt : %s.xe;\n" ctx ctx ctx ctx *)
    Soc2cUtil.gen_assign t (Printf.sprintf "%s.z" ctx)
       (Printf.sprintf "(%s.c)? %s.xt : %s.xe" ctx ctx ctx)
       (Printf.sprintf "sizeof(%s.z)" ctx)

let (lustre_impl : Soc.key -> string) =
  fun sk -> 
    let ctx = get_ctx_name sk in
    (* use gen_assign? *)
    Printf.sprintf"  %s.z = (!%s.x || %s.y);\n" ctx ctx ctx 

let (lustre_arrow : Soc.key -> string) =
  fun sk -> 
    let x,y,z = "ctx->x", "ctx->y", "ctx->z" in
    let (_,t::_,_) = sk in
    let vo = Printf.sprintf"((ctx->_memory)? %s : %s)" x y in
    let size = Printf.sprintf "sizeof(%s)" x in
    (Soc2cUtil.gen_assign t z vo size) ^ 
      ("  ctx->_memory = _false;\n")
      
let (lustre_merge : Soc.key -> string) =
  fun sk -> 
    let ctx = get_ctx_name sk in
    let (_,tc::tl,_) = sk in
    match tc with
      | Bool ->
        Printf.sprintf"  %s.z = (%s.clk) ? %s.x1 : %s.x2  ;\n" ctx ctx ctx ctx
      | Enum(en,el) ->
        let case_list = List.mapi
          (fun i e -> Printf.sprintf "    case %s: %s.x%i; break;\n" (id2s e) ctx i) el  
        in
        let cases = String.concat "" case_list in
        Printf.sprintf"  %s.z =\n    switch (%s.clk){\n%s}\n" ctx ctx cases
      | Int ->
        let case_list = List.mapi
          (fun i e -> Printf.sprintf "    case %i: %s.z = %s.x%i; break;\n" i ctx ctx i) tl  
        in
        let cases = String.concat "" case_list in
        Printf.sprintf"  switch (%s.clk){\n%s}\n" ctx cases
      | _ -> assert false


let (lustre_hat : Soc.key -> string) =
  fun (n,tl,si_opt) -> 
    let ctx = get_ctx_name (n,tl,si_opt) in
    let i,t = match tl with
      | [_;Data.Array(t,i)] -> i,t
      | _ -> assert false
    in
    let buff = ref "" in
    for j=0 to i-1 do
      buff := !buff^(Soc2cUtil.gen_assign t (Printf.sprintf "%s.z[%d]" ctx j)
                       (Printf.sprintf "%s.x" ctx) (Printf.sprintf "sizeof(%s.x)" ctx)); 
    done;
    !buff

let (lustre_array: Soc.key -> string) =
  fun (n,tl,si_opt) -> 
    let ctx = get_ctx_name (n,tl,si_opt) in
    let t,i = match List.hd (List.rev tl) with
      | Data.Array(t,i) -> t,i
      | _ -> assert false
    in
    let buff = ref "" in
    for j=0 to i-1 do
      buff := !buff^(Soc2cUtil.gen_assign t (Printf.sprintf "%s.z[%d]" ctx j)
                       (Printf.sprintf "%s.x%d" ctx (j+1)) 
                       (Printf.sprintf "sizeof(%s.x%d)" ctx (j+1))); 
    done;
    !buff
 
let (lustre_concat: Soc.key -> string) =
  fun (n,tl,si_opt) -> 
    let ctx = get_ctx_name (n,tl,si_opt) in
    let t,s1,s2 = match tl with
      | [Data.Array(t,s1); Data.Array(_,s2); _] -> t,s1,s2
      | _ -> assert false
    in
    let t1 = Printf.sprintf "%s.x" ctx
    and t2 = Printf.sprintf "%s.y" ctx  
    and t12 = Printf.sprintf "%s.z" ctx in  
    (Printf.sprintf "  memcpy(%s, %s, sizeof(%s));\n" t12 t1 t1)^ 
      (Printf.sprintf "  memcpy(%s[%d], %s, sizeof(%s));\n" t12 s1 t2 t2) 
(* Both seems to work *)
(*      let buff = ref "" in  *)
(*      for j=0 to s1-1 do  *)
(*        buff := !buff^(Soc2cUtil.gen_assign t (Printf.sprintf "%s.z[%d]" ctx j)  *)
(*                         (Printf.sprintf "%s.x[%d]" ctx j));   *)
(*      done;  *)
(*      for j=s1 to s1+s2-1 do  *)
(*        buff := !buff^(Soc2cUtil.gen_assign t (Printf.sprintf "%s.z[%d]" ctx j)  *)
(*                         (Printf.sprintf "%s.y[%d]" ctx (j-s1)));   *)
(*      done;  *)
(*      !buff  *)

let (lustre_slice: Soc.key -> string) =
  fun (n,tl,si_opt) -> 
    let ctx = get_ctx_name (n,tl,si_opt) in
    let t,size = match List.hd (List.rev tl) with
      | Data.Array(t,i) -> t,i
      | _ -> assert false
    in
    match si_opt with
      | Slic(b,e,step) -> 
        let buff = ref "" in  
        let j=ref 0 in
        for i = b to e do
          if (i-b) mod step = 0 then (
            buff := !buff^(Soc2cUtil.gen_assign t (Printf.sprintf "%s.z[%d]" ctx !j)  
                             (Printf.sprintf "%s.x[%d]" ctx i)
                             (Printf.sprintf "sizeof(%s.x[%d])" ctx i)
            );   
            incr j);
          done; 
          !buff
      | _ -> assert false



(* exported *)
let (get_key: Soc.key -> string) =
  fun sk -> 
    let (n,tl,si_opt) = sk in
    match n with
      | "Lustre::rplus" 
      | "Lustre::plus"
      | "Lustre::iplus"  -> lustre_binop sk "+" 
      | "Lustre::itimes" 
      | "Lustre::times"
      | "Lustre::rtimes" -> lustre_binop sk "*" 
      | "Lustre::idiv"
      | "Lustre::div"  
      | "Lustre::rdiv"   -> lustre_binop sk "/" 
      | "Lustre::islash" 
      | "Lustre::slash" 
      | "Lustre::rslash" -> lustre_binop sk "/"
      | "Lustre::iminus"
      | "Lustre::minus"
      | "Lustre::rminus" -> lustre_binop sk "-" 

      | "Lustre::mod" -> lustre_binop sk "%"
      | "Lustre::iuminus"
      | "Lustre::uminus" 
      | "Lustre::ruminus"-> lustre_unop sk "-" 

      | "Lustre::eq" -> lustre_binop sk "==" 

      | "Lustre::and" -> lustre_binop sk "&&" 
      | "Lustre::neq" -> lustre_binop sk "!="  
      | "Lustre::or"  -> lustre_binop sk "||"  

      | "Lustre::xor" -> lustre_binop sk "^" 
      | "Lustre::not" -> lustre_unop  sk "!"
      | "Lustre::real2int" -> lustre_unop sk "(_integer)"
      | "Lustre::int2real" -> lustre_unop sk "(_real)"

      | "Lustre::lt"
      | "Lustre::rlt"
      | "Lustre::ilt" -> lustre_binop sk "<"
      | "Lustre::gt"
      | "Lustre::rgt"
      | "Lustre::igt" -> lustre_binop sk ">"
      | "Lustre::lte" 
      | "Lustre::rlte"
      | "Lustre::ilte" -> lustre_binop sk "<="
      | "Lustre::gte"
      | "Lustre::rgte"
      | "Lustre::igte" -> lustre_binop sk ">="
      | "Lustre::impl" -> lustre_impl sk

      | "Lustre::if"
      | "Lustre::rif"
      | "Lustre::iif" -> lustre_ite sk

      | "Lustre::arrow" -> lustre_arrow sk
      | "Lustre::merge" -> lustre_merge sk

      | "Lustre::hat" -> lustre_hat sk
      | "Lustre::array" -> lustre_array sk
      | "Lustre::concat" -> lustre_concat sk
      | "Lustre::array_slice" -> lustre_slice sk

      | "Lustre::current" -> assert false (* o*)
      | "Lustre::nor" -> assert false (* ougth to be translated into boolred *)
      | "Lustre::diese" -> assert false (* ditto *)

      | _ -> assert false


let rec type_elt_of_array = function 
  | Data.Array(t,_) -> t
  | Data.Alias(_,t) -> type_elt_of_array t
  | _ -> assert false


(* exported *)
let (get_iterator : Soc.t -> string -> Soc.t -> int -> string) = 
  fun soc iterator it_soc n -> 
    let iter_inputs,iter_outputs = soc.profile in
    let node_step = match soc.step with [step] -> step.name  | _ ->  assert false in
    let step_args, ctx, array_index,ctx_access = 
      match soc.instances with
        | [] -> (
          let ctx_access = Printf.sprintf "%s." (get_ctx_name soc.key) in
          let (array_index : int -> var -> Soc.var_expr) =
            fun i (vn,vt) -> Var(Printf.sprintf "%s%s[%d]" ctx_access vn i, type_elt_of_array vt)
          in          
          Array.make n "", 
          Array.make n (get_ctx_name it_soc.key),
          array_index,ctx_access
        )
        | _  -> 
          let inst_names = List.map fst soc.instances in 
          let inst_names = List.rev inst_names in
          let step_args = List.map (fun sn  -> ("&ctx->"^(id2s sn))) inst_names in
          let ctx = List.map (fun sn  -> ("ctx->"^(id2s sn))) inst_names in
          let ctx_access =  "ctx->"  in
          let (array_index : int -> var -> Soc.var_expr) =
            fun i (vn,vt) -> Var(Printf.sprintf "ctx->%s[%d]" vn i,vt) 
          in
          Array.of_list step_args,
          Array.of_list ctx,
          array_index,ctx_access
    in
    let buff = ref "" in
    for i=0 to n-1 do
      let vel_in, vel_out =
        match iterator with
          | "map" -> 
            (List.map (array_index i) iter_inputs,
             List.map (array_index i) iter_outputs)
          | "fold" | "red" | "fill" | "fillred" ->
            let name, telt = List.hd iter_inputs in
            let a_in = ctx_access ^ name in
            let a_in = Var(a_in, telt) in
            ( a_in::(List.map (array_index i) (List.tl iter_inputs)),
              a_in::(List.map (array_index i) (List.tl iter_outputs)))
          | _ -> assert false (* should not occur *)
      in
      buff := !buff^(
        Soc2cUtil.gen_step_call
          soc it_soc vel_out vel_in ctx.(i) node_step step_args.(i))
    done;

    if iterator <> "map" then (
      let type_in = (snd (List.hd iter_inputs)) in
      let a_in  = ctx_access ^ (fst (List.hd iter_inputs)) in
      let a_out = ctx_access ^ (fst (List.hd iter_outputs)) in
      buff := !buff^(Soc2cUtil.gen_assign type_in a_out a_in
                       (Printf.sprintf "sizeof(%s)" a_in))  (* a_out=a_n *)
    );
    !buff
 

(* exported *)
let (get_condact : Soc.t -> Soc.t -> var_expr list -> string ) = 
  fun soc condact_soc el -> 
    let ctx = get_ctx_name soc.key in
    let buff = ref "" in
    let add str = buff:=!buff^(str^"\n") in

    let clk = Printf.sprintf "ctx->i0" in
    let string_of_var_soc (id,_) = "ctx->"^(id2s id) in
    let vel_in,vel_out = soc.profile in
    let vel_in = List.tl vel_in in
    let vel_in  = List.map (fun var -> Var var) vel_in  in
    let vel_out = List.map (fun var -> Var var) vel_out in
    add (Printf.sprintf "  if (%s == _true) { " clk); 
    if SocUtils.is_memory_less condact_soc then
      let condact_ctx = get_ctx_name condact_soc.key in
      add (Soc2cUtil.gen_step_call soc condact_soc vel_out vel_in condact_ctx "step" "")
    else (
      let condact_ctx = 
        let inst_name =
          match soc.instances with
            | [inst] -> (id2s (fst inst))
            | _ -> assert false
        in
        Printf.sprintf  "ctx->%s" inst_name
      in
      add (Soc2cUtil.gen_step_call soc condact_soc vel_out vel_in 
             condact_ctx "step" ("&"^condact_ctx))
    );
    add "    ctx->_memory = _false;";
    add "   } else if (ctx->_memory == _true) {";
    List.iter2 (fun var ve -> 
      add (Printf.sprintf "    %s = %s;" (Soc2cUtil.string_of_var_expr soc var)
             (Soc2cUtil.string_of_var_expr soc ve) )
    ) vel_out el ;
    add "    ctx->_memory = _false;";
    add "  }";
    !buff


(* exported *)
let (get_boolred : Soc.t -> int -> int -> int -> string )= 
  fun soc i j k -> 
    let ctx = get_ctx_name soc.key in
    let buff = ref "" in
    let add str = buff:=!buff^(str^"\n") in
    add "  int cpt,i;";
    add "  cpt=0;";
    add (Printf.sprintf
           "  for (i = 0 ; i < %d ; i += 1) if (%s.i0[i] == _true) cpt++;" k ctx);    
    add (Printf.sprintf "  %s.o0 = (%d <= cpt && cpt <= %d) ? _true : _false;" ctx i j );
    !buff
